@@ -493,11 +493,11 @@ static void save_attachment_dialog(void)
 		{
 			fz_try(ctx)
 			{
-				pdf_obj *fs = pdf_dict_get(ctx, pdf_annot_obj(ctx, ui.selected_annot), PDF_NAME(FS));
+				pdf_obj *fs = pdf_annot_filespec(ctx, ui.selected_annot);
 				fz_buffer *buf = pdf_load_embedded_file_contents(ctx, fs);
 				fz_save_buffer(ctx, buf, attach_filename);
 				fz_drop_buffer(ctx, buf);
-				trace_action("tmp = annot.getFileSpec()\n");
+				trace_action("tmp = annot.getFilespec()\n");
 				trace_action("doc.getEmbeddedFileContents(tmp).save(\"%s\");\n", attach_filename);
 				trace_action("tmp = doc.verifyEmbeddedFileChecksum(tmp);\n");
 				trace_action("if (tmp != true)\n");
@@ -537,7 +537,7 @@ static void open_attachment_dialog(void)
 					created, modified, 0);
 				pdf_set_annot_filespec(ctx, ui.selected_annot, fs);
 				fz_drop_buffer(ctx, contents);
-				trace_action("annot.setFileSpec(doc.addEmbeddedFile(\"%s\", null, readFile(\"%s\"), new Date(%d).getTime(), new Date(%d).getTime(), false));\n", filename, attach_filename, created, modified);
+				trace_action("annot.setFilespec(doc.addEmbeddedFile(\"%s\", null, readFile(\"%s\"), new Date(%d).getTime(), new Date(%d).getTime(), false));\n", filename, attach_filename, created, modified);
 			}
 			fz_always(ctx)
 			{
@@ -1474,7 +1474,7 @@ void do_annotate_panel(void)
 
 		if (pdf_annot_type(ctx, ui.selected_annot) == PDF_ANNOT_FILE_ATTACHMENT)
 		{
-			pdf_embedded_file_params params;
+			pdf_filespec_params params;
 			char attname[PATH_MAX];
 			pdf_obj *fs = pdf_annot_filespec(ctx, ui.selected_annot);
 			ui_spacer();
@@ -1484,7 +1484,7 @@ void do_annotate_panel(void)
 				{
 					fz_dirname(attname, filename, sizeof attname);
 					fz_strlcat(attname, "/", sizeof attname);
-					pdf_get_embedded_file_params(ctx, fs, &params);
+					pdf_get_filespec_params(ctx, fs, &params);
 					fz_strlcat(attname, params.filename, sizeof attname);
 					ui_init_save_file(attname, NULL);
 					ui.dialog = save_attachment_dialog;
