@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -68,9 +68,15 @@ typedef enum
  *	cmd:	What type of node this is.
  *
  *	size:	The number of sizeof(fz_display_node) bytes that this node's
- *		data occupies. (i.e. &node[node->size] = the next node in the
+ *		data occupies. (i.e. &node[size] = the next node in the
  *		chain; 0 for end of list).
  *
+ *		At 9 bits for this field, and 4 bytes for a node, this means
+ *		the largest node can span 511*4 bytes. We therefore reserve
+ *		511 as a special value to mean that the actual size for the
+ *		node is given in an (unaligned!) size_t following the node
+ *		before the rest of the data.
+ *
  *	rect:	0 for unchanged, 1 for present.
  *
  *	path:	0 for unchanged, 1 for present.
@@ -135,7 +141,7 @@ enum {
 	CTM_CHANGE_BC = 2,
 	CTM_CHANGE_EF = 4,
 
-	MAX_NODE_SIZE = (1<<9)-sizeof(fz_display_node)
+	INDIRECT_NODE_THRESHOLD = (1<<9)-1
 };
 
 struct fz_display_list
@@ -211,6 +217,22 @@ static void align_node_for_pointer(fz_display_node **node)
 		(*node) = (fz_display_node *)((ptr + FZ_POINTER_ALIGN_MOD - 1) & ~(FZ_POINTER_ALIGN_MOD-1));
 }
 
+static int
+cmd_needs_alignment(fz_display_command cmd)
+{
+	return (cmd == FZ_CMD_FILL_TEXT ||
+		cmd == FZ_CMD_STROKE_TEXT ||
+		cmd == FZ_CMD_CLIP_TEXT ||
+		cmd == FZ_CMD_CLIP_STROKE_TEXT ||
+		cmd == FZ_CMD_IGNORE_TEXT ||
+		cmd == FZ_CMD_FILL_SHADE ||
+		cmd == FZ_CMD_FILL_IMAGE ||
+		cmd == FZ_CMD_FILL_IMAGE_MASK ||
+		cmd == FZ_CMD_CLIP_IMAGE_MASK ||
+		cmd == FZ_CMD_END_MASK ||
+		cmd == FZ_CMD_DEFAULT_COLORSPACES);
+}
+
 static unsigned char *
 fz_append_display_node(
 	fz_context *ctx,
@@ -327,7 +349,7 @@ fz_append_display_node(
 	if (color == NULL)
 	{
 		if (colorspace)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Colorspace cannot be specified without color.");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Colorspace cannot be specified without color.");
 	}
 	else
 	{
@@ -510,7 +532,7 @@ fz_append_display_node(
 			ctm_flags |= CTM_CHANGE_EF, size += SIZE_IN_NODES(2*sizeof(float));
 		node.ctm = ctm_flags;
 	}
-	if (stroke && (writer->stroke == NULL || stroke != writer->stroke))
+	if (stroke && (writer->stroke == NULL || !fz_stroke_state_eq(ctx, stroke, writer->stroke)))
 	{
 		pad_size_for_pointer(list, &size);
 		stroke_off = size;
@@ -519,11 +541,8 @@ fz_append_display_node(
 	}
 	if (path && (writer->path == NULL || path != writer->path))
 	{
-		size_t max;
-
 		pad_size_for_pointer(list, &size);
-		max = SIZE_IN_NODES(MAX_NODE_SIZE) - size - SIZE_IN_NODES(private_data_len);
-		path_size = SIZE_IN_NODES(fz_pack_path(ctx, NULL, max, path));
+		path_size = SIZE_IN_NODES(fz_pack_path(ctx, NULL, path));
 		node.path = 1;
 		path_off = size;
 
@@ -531,16 +550,18 @@ fz_append_display_node(
 	}
 	if (private_data_len)
 	{
-		size_t max;
-
-		pad_size_for_pointer(list, &size);
-		max = SIZE_IN_NODES(MAX_NODE_SIZE) - size;
-		if (SIZE_IN_NODES(private_data_len) > max)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Private data too large to pack into display list node");
+		if (cmd_needs_alignment(cmd))
+			pad_size_for_pointer(list, &size);
 		private_off = size;
 		size += SIZE_IN_NODES(private_data_len);
 	}
 
+	/* If the size is more than 511, then we can't signal that in 9 bits,
+	 * so we'll send it as 511, and then put an extra size_t with the
+	 * size in. */
+	if (size >= INDIRECT_NODE_THRESHOLD)
+		size += SIZE_IN_NODES(sizeof(size_t));
+
 	while (list->len + size > list->max)
 	{
 		size_t newsize = list->max * 2;
@@ -563,29 +584,35 @@ fz_append_display_node(
 			writer->path = (fz_path *)(((char *)writer->path) + diff);
 	}
 
-	if ((unsigned int)size != size)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Display list node too large");
-
 	/* Write the node to the list */
-	node.size = (unsigned int)size;
+	if (size >= INDIRECT_NODE_THRESHOLD)
+		node.size = INDIRECT_NODE_THRESHOLD;
+	else
+		node.size = (unsigned int)size;
 	node.flags = flags;
-	assert(size < (1<<9));
 	node_ptr = &list->list[list->len];
 	*node_ptr = node;
 
+	/* Insert the explicit size (unaligned) if required. */
+	if (size >= INDIRECT_NODE_THRESHOLD)
+	{
+		memcpy(&node_ptr[1], &size, sizeof(size));
+		node_ptr += SIZE_IN_NODES(sizeof(size_t));
+	}
+
 	/* Path is the most frequent one, so try to avoid the try/catch in
 	 * this case */
 	if (path_off)
 	{
 		my_path = (void *)(&node_ptr[path_off]);
-		(void)fz_pack_path(ctx, (void *)my_path, path_size * sizeof(fz_display_node), path);
+		(void)fz_pack_path(ctx, (void *)my_path, path);
 	}
 
 	if (stroke_off)
 	{
 		fz_try(ctx)
 		{
-			my_stroke = fz_keep_stroke_state(ctx, stroke);
+			my_stroke = fz_clone_stroke_state(ctx, stroke);
 		}
 		fz_catch(ctx)
 		{
@@ -700,7 +727,7 @@ fz_append_display_node(
 		fz_stroke_state **out_stroke = (fz_stroke_state **)(void *)(&node_ptr[stroke_off]);
 		*out_stroke = my_stroke;
 		fz_drop_stroke_state(ctx, writer->stroke);
-		/* Can never fail as my_stroke was kept above */
+		/* Can never fail as my_stroke was cloned above */
 		writer->stroke = fz_keep_stroke_state(ctx, my_stroke);
 	}
 	if (private_data_len)
@@ -1127,30 +1154,37 @@ fz_list_begin_mask(fz_context *ctx, fz_device *dev, fz_rect rect, int luminosity
 }
 
 static void
-fz_list_end_mask(fz_context *ctx, fz_device *dev)
+fz_list_end_mask(fz_context *ctx, fz_device *dev, fz_function *tr)
 {
-	fz_append_display_node(
-		ctx,
-		dev,
-		FZ_CMD_END_MASK,
-		0, /* flags */
-		NULL, /* rect */
-		NULL, /* path */
-		NULL, /* color */
-		NULL, /* colorspace */
-		NULL, /* alpha */
-		NULL, /* ctm */
-		NULL, /* stroke */
-		NULL, /* private_data */
-		0); /* private_data_len */
+	fz_function *tr2 = fz_keep_function(ctx, tr);
+
+	fz_try(ctx)
+		fz_append_display_node(
+			ctx,
+			dev,
+			FZ_CMD_END_MASK,
+			0, /* flags */
+			NULL, /* rect */
+			NULL, /* path */
+			NULL, /* color */
+			NULL, /* colorspace */
+			NULL, /* alpha */
+			NULL, /* ctm */
+			NULL, /* stroke */
+			&tr2, /* private_data */
+			sizeof(tr2)); /* private_data_len */
+	fz_catch(ctx)
+	{
+		fz_drop_function(ctx, tr);
+		fz_rethrow(ctx);
+	}
 }
 
 static void
 fz_list_begin_group(fz_context *ctx, fz_device *dev, fz_rect rect, fz_colorspace *colorspace, int isolated, int knockout, int blendmode, float alpha)
 {
 	int flags;
-
-	colorspace = fz_keep_colorspace(ctx, colorspace);
+	static const float color[FZ_MAX_COLORS] = { 0 };
 
 	flags = (blendmode<<2);
 	if (isolated)
@@ -1158,28 +1192,20 @@ fz_list_begin_group(fz_context *ctx, fz_device *dev, fz_rect rect, fz_colorspace
 	if (knockout)
 		flags |= KNOCKOUT;
 
-	fz_try(ctx)
-	{
-		fz_append_display_node(
-			ctx,
-			dev,
-			FZ_CMD_BEGIN_GROUP,
-			flags,
-			&rect,
-			NULL, /* path */
-			NULL, /* color */
-			NULL, /* colorspace */
-			&alpha, /* alpha */
-			NULL, /* ctm */
-			NULL, /* stroke */
-			&colorspace, /* private_data */
-			sizeof(colorspace)); /* private_data_len */
-	}
-	fz_catch(ctx)
-	{
-		fz_drop_colorspace(ctx, colorspace);
-		fz_rethrow(ctx);
-	}
+	fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_BEGIN_GROUP,
+		flags,
+		&rect,
+		NULL, /* path */
+		color, /* color */
+		colorspace, /* colorspace */
+		&alpha, /* alpha */
+		NULL, /* ctm */
+		NULL, /* stroke */
+		NULL, /* private_data */
+		0); /* private_data_len */
 }
 
 static void
@@ -1318,6 +1344,8 @@ fz_list_set_default_colorspaces(fz_context *ctx, fz_device *dev, fz_default_colo
 static void
 fz_list_begin_layer(fz_context *ctx, fz_device *dev, const char *layer_name)
 {
+	size_t len = layer_name ? strlen(layer_name) : 0;
+
 	fz_append_display_node(
 		ctx,
 		dev,
@@ -1330,8 +1358,8 @@ fz_list_begin_layer(fz_context *ctx, fz_device *dev, const char *layer_name)
 		NULL, /* alpha */
 		NULL,
 		NULL, /* stroke */
-		layer_name, /* private_data */
-		1+strlen(layer_name)); /* private_data_len */
+		len ? layer_name : "", /* private_data */
+		len + 1); /* private_data_len */
 }
 
 static void
@@ -1354,7 +1382,7 @@ fz_list_end_layer(fz_context *ctx, fz_device *dev)
 }
 
 static void
-fz_list_begin_structure(fz_context *ctx, fz_device *dev, fz_structure standard, const char *raw, int uid)
+fz_list_begin_structure(fz_context *ctx, fz_device *dev, fz_structure standard, const char *raw, int idx)
 {
 	unsigned char *data;
 	size_t len = (raw ? strlen(raw) : 0);
@@ -1372,13 +1400,13 @@ fz_list_begin_structure(fz_context *ctx, fz_device *dev, fz_structure standard,
 		NULL,
 		NULL, /* stroke */
 		NULL, /* private_data */
-		len+2+sizeof(uid)); /* private_data_len */
+		len+2+sizeof(idx)); /* private_data_len */
 	data[0] = (char)standard;
-	memcpy(data+1, &uid, sizeof(uid));
+	memcpy(data+1, &idx, sizeof(idx));
 	if (len)
-		memcpy(data+1+sizeof(uid), raw, len+1);
+		memcpy(data+1+sizeof(idx), raw, len+1);
 	else
-		data[1+sizeof(uid)] = 0;
+		data[1+sizeof(idx)] = 0;
 }
 
 static void
@@ -1529,7 +1557,17 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 	while (node != node_end)
 	{
 		fz_display_node n = *node;
-		fz_display_node *next = node + n.size;
+		size_t size = n.size;
+		fz_display_node *next;
+
+		if (size == INDIRECT_NODE_THRESHOLD)
+		{
+			memcpy(&size, &node[1], sizeof(size));
+			node += SIZE_IN_NODES(sizeof(size_t));
+			size -= SIZE_IN_NODES(sizeof(size_t));
+		}
+
+		next = node + size;
 
 		node++;
 		if (n.rect)
@@ -1609,9 +1647,9 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 			align_node_for_pointer(&node);
 			fz_drop_image(ctx, *(fz_image **)node);
 			break;
-		case FZ_CMD_BEGIN_GROUP:
+		case FZ_CMD_END_MASK:
 			align_node_for_pointer(&node);
-			fz_drop_colorspace(ctx, *(fz_colorspace **)node);
+			fz_drop_function(ctx, *(fz_function **)node);
 			break;
 		case FZ_CMD_DEFAULT_COLORSPACES:
 			align_node_for_pointer(&node);
@@ -1700,8 +1738,16 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 	{
 		int empty;
 		fz_display_node n = *node;
+		size_t size = n.size;
 
-		next_node = node + n.size;
+		if (size == INDIRECT_NODE_THRESHOLD)
+		{
+			memcpy(&size, &node[1], sizeof(size_t));
+			node += SIZE_IN_NODES(sizeof(size_t));
+			size -= SIZE_IN_NODES(sizeof(size_t));
+		}
+
+		next_node = node + size;
 
 		/* Check the cookie for aborting */
 		if (cookie)
@@ -1709,7 +1755,7 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 			if (cookie->abort)
 				break;
 			cookie->progress = progress;
-			progress += n.size;
+			progress += (int)size;
 		}
 
 		node++;
@@ -1872,6 +1918,10 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 			empty = fz_is_empty_rect(fz_intersect_rect(trans_rect, scissor));
 		}
 
+		/* clipped starts out as 0. It only goes non-zero here if we move inside
+		 * an 'empty' region. Whenever clipped is non zero, or we are in an empty
+		 * region, we therefore may need to increment clipped according to the
+		 * nesting. */
 		if (clipped || empty)
 		{
 			switch (n.cmd)
@@ -1885,6 +1935,13 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 			case FZ_CMD_BEGIN_GROUP:
 				clipped++;
 				continue;
+			case FZ_CMD_BEGIN_STRUCTURE:
+			case FZ_CMD_END_STRUCTURE:
+			case FZ_CMD_BEGIN_METATEXT:
+			case FZ_CMD_END_METATEXT:
+				/* These may not nest as nicely as we'd like. Just ignore them for
+				 * the purposes of clipping. */
+				break;
 			case FZ_CMD_POP_CLIP:
 			case FZ_CMD_END_GROUP:
 				if (!clipped)
@@ -1970,11 +2027,11 @@ visible:
 				fz_begin_mask(ctx, dev, trans_rect, n.flags & 1, colorspace, color, color_params);
 				break;
 			case FZ_CMD_END_MASK:
-				fz_end_mask(ctx, dev);
+				align_node_for_pointer(&node);
+				fz_end_mask_tr(ctx, dev, *(fz_function **)node);
 				break;
 			case FZ_CMD_BEGIN_GROUP:
-				align_node_for_pointer(&node);
-				fz_begin_group(ctx, dev, trans_rect, *(fz_colorspace **)node, (n.flags & ISOLATED) != 0, (n.flags & KNOCKOUT) != 0, (n.flags>>2), alpha);
+				fz_begin_group(ctx, dev, trans_rect, colorspace, (n.flags & ISOLATED) != 0, (n.flags & KNOCKOUT) != 0, (n.flags>>2), alpha);
 				break;
 			case FZ_CMD_END_GROUP:
 				fz_end_group(ctx, dev);
@@ -1984,7 +2041,6 @@ visible:
 				int cached;
 				fz_list_tile_data *data;
 				fz_rect tile_rect;
-				align_node_for_pointer(&node);
 				data = (fz_list_tile_data *)node;
 				tiled++;
 				tile_rect = data->view;
@@ -2008,7 +2064,6 @@ visible:
 				fz_set_default_colorspaces(ctx, dev, *(fz_default_colorspaces **)node);
 				break;
 			case FZ_CMD_BEGIN_LAYER:
-				align_node_for_pointer(&node);
 				fz_begin_layer(ctx, dev, (const char *)node);
 				break;
 			case FZ_CMD_END_LAYER:
@@ -2017,11 +2072,10 @@ visible:
 			case FZ_CMD_BEGIN_STRUCTURE:
 			{
 				const unsigned char *data;
-				int uid;
-				align_node_for_pointer(&node);
+				int idx;
 				data = (const unsigned char *)node;
-				memcpy(&uid, data+1, sizeof(uid));
-				fz_begin_structure(ctx, dev, (fz_structure)data[0], (const char *)(data[1+sizeof(uid)] == 0 ? NULL : &data[1+sizeof(uid)]), uid);
+				memcpy(&idx, data+1, sizeof(idx));
+				fz_begin_structure(ctx, dev, (fz_structure)data[0], (const char *)(&data[1+sizeof(idx)]), idx);
 				break;
 			}
 			case FZ_CMD_END_STRUCTURE:
@@ -2030,9 +2084,10 @@ visible:
 			case FZ_CMD_BEGIN_METATEXT:
 			{
 				const unsigned char *data;
-				align_node_for_pointer(&node);
+				const char *text;
 				data = (const unsigned char *)node;
-				fz_begin_metatext(ctx, dev, (fz_metatext)data[0], (const char *)(data[1] == 0 ? NULL : &data[1]));
+				text = (const char *)&data[1];
+				fz_begin_metatext(ctx, dev, (fz_metatext)data[0], text);
 				break;
 			}
 			case FZ_CMD_END_METATEXT:
@@ -2042,11 +2097,22 @@ visible:
 		}
 		fz_catch(ctx)
 		{
+			if (fz_caught(ctx) == FZ_ERROR_SYSTEM)
+			{
+				fz_drop_colorspace(ctx, colorspace);
+				fz_drop_stroke_state(ctx, stroke);
+				fz_drop_path(ctx, path);
+				fz_rethrow(ctx);
+			}
 			/* Swallow the error */
 			if (cookie)
 				cookie->errors++;
 			if (fz_caught(ctx) == FZ_ERROR_ABORT)
+			{
+				fz_ignore_error(ctx);
 				break;
+			}
+			fz_report_error(ctx);
 			fz_warn(ctx, "Ignoring error during interpretation");
 		}
 	}
